/*
 * ejabberd external authentication program
 *
 * (c) Alexander Vdolainen 2013, 2018, 2019, 2021 <alex@vapaa.xyz>
 *
 * this is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * this is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 *
 */

#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <errno.h>
#include <sys/uio.h>
#include <string.h>

#include <ejabbermsg.h>

ssize_t eport_read(FILE *stream, char *buffer, size_t buffer_size)
{
  int fd = fileno(stream);
  int16_t port_msg_len = 0;
  ssize_t pr = 0;

  if(fd < 0) return -1;

  /* firstly the size of plain text message is required.
   * it has a short size and is going in bigendian order.
   */
  pr = read(fd, &port_msg_len, sizeof(int16_t));
  if(pr < 0) return -1;
  port_msg_len = (port_msg_len << 8) | ((port_msg_len >> 8) & 0xFF);

  /* check if a buffer has required amount of bytes */
  if(port_msg_len > buffer_size) {
    errno = EMSGSIZE; /* message is too long */
    return -1;
  } else
    pr = read(fd, buffer, port_msg_len);

  return pr;
}

ssize_t eport_write(FILE *stream, char *buffer, size_t buffer_size)
{
  int fd = fileno(stream);
  int16_t port_msg_len = (buffer_size << 8) | ((buffer_size >> 8) & 0xFF);
  ssize_t wr = 0;
  struct iovec msg[2];

  if(buffer_size <= 0) {
    errno = EINVAL;
    return -1;
  }

  /* write length and message itself */
  msg[0].iov_base = &port_msg_len;
  msg[0].iov_len = sizeof(int16_t);
  msg[1].iov_base = buffer;
  msg[1].iov_len = buffer_size;

  wr = writev(fd, msg, 2);

  return wr;
}

int eport_ejabberd_msgread(char *buffer, size_t buffer_size,
                           struct ejabber_msg *m)
{
  char *orig_buffer = buffer;
  unsigned long hash = 5381;
  int c, t = 0;

  if(!buffer || buffer_size <= 0) {
  ieinval:
    errno = EINVAL;
    return -1;
  }
  if(!m) goto ieinval;

  /* calc hash for first message word till the end or till the ':' */
  while(((c = *buffer++) != '\0') && (c != ':'))
    hash = ((hash << 5) + hash) + c;

  switch(hash) {
  case AUTH_HASH: m->aev = EJA_AUTH; break;
  case ISUSER_HASH: m->aev = EJA_ISUSER; break;
  case SETPASS_HASH: m->aev = EJA_SETPASS; break;
  case TRYREGISTER_HASH: m->aev = EJA_TRYREGISTER; break;
  case REMOVEUSER_HASH: m->aev = EJA_REMOVEUSER; break;
  case REMOVEUSER3_HASH: m->aev = EJA_REMOVEUSER3; break;
  default:
    errno = EINVAL;
    return -1;
  }
  /* initialize other things in message */
  m->msg_data = orig_buffer;
  m->user = m->domain = m->password = NULL;
  /* shift a pointer */
  c = *buffer; orig_buffer = buffer;

  while((buffer = strchr(buffer, ':')) != NULL) {
    switch(t) {
    case 0: m->user = orig_buffer; break;
    case 1: m->domain = orig_buffer; break;
    default:
      errno = EINVAL;
      return -1;
    }
    *buffer = '\0'; c = *buffer++;
    orig_buffer = buffer; t++;
  }
  if(*orig_buffer != '\0') {
    if(t > 1) m->password = orig_buffer;
    else m->domain = orig_buffer;
    t++;
  }

  return t;
}
