/*
 * ejabberd external authentication program
 *
 * (c) Alexander Vdolainen 2013, 2018, 2019, 2021 <alex@vapaa.xyz>
 *
 * this is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * this is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/select.h>
#include <errno.h>
#include <netdb.h>
#include <string.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <stdarg.h>
#include <getopt.h>

#include <ejabbermsg.h>
#include <tlsport.h>
#include <smtpf.h>

#define PRG_NAME  "ejabberauth smtp function test"

extern int ejda_username_split(char *buf, char **user, char **domain);

/*
 * The idea is stupid simple: take hostname, port, username and password
 * move this data to the jabber message manually, connect to the host,
 * run the function, check result and disconnect.
 * Extra option -e (--expected-result) ok/fail to check the output from
 * imap according to the expectations
 */

static void print_help(const char *cname)
{
  fprintf(stdout, "\n%s - %s\n\n", cname, PRG_NAME);
  fprintf(stdout, "USAGE:\n");
  fprintf(stdout, "\t%s -H | --host <SMTP hostname> -p | --port <port> -u | --user <username>"
          " [ -e | --expected-result <ok | fail>] [ -v | --verbose ]", cname);
  fprintf(stdout, "\n\n\t%s -h | --help\n", cname);

  return;
}

int main(int argc, char **argv)
{
  char *host, *port, *user;
  char *utbuf = NULL;
  int verbose = 0, opt, expected = 0;
  struct ejabber_msg m;
  struct tlsport p;

  host = port = user = NULL;

  /* parse options */
  while(1) {
    int option_index = 0;
    static struct option long_options[] = {
      {"help", no_argument, NULL, 'h'},
      {"host", required_argument, NULL, 'H'},
      {"port", required_argument, NULL, 'p'},
      {"user", required_argument, NULL, 'u'},
      {"expected-result", required_argument, NULL, 'e'},
      {"verbose", no_argument, NULL, 'v'},
      {NULL, 0, NULL, 0},
    };

    if((opt = getopt_long(argc, argv, "hH:p:u:e:v", long_options,
                          &option_index)) == -1) break;
    switch(opt) {
    case 'h':
      /* print help information and exit */
      print_help(argv[0]);
      return 0;
      break;
    case 'H': host = optarg; break;
    case 'p': port = optarg; break;
    case 'u': user = optarg; break;
    case 'e':
      if(!strcmp(optarg, "fail")) expected = 1;
      else expected = 0;
      break;
    case 'v': verbose = 1; break;
    default: abort(); break; /* something that's shouldn't ever happen */
    }
  }

  if(!host) {
    fprintf(stderr, "Error: no hostname given.\n");
  main_einval:
    print_help(argv[0]);
    return EINVAL;
  }
  if(!port) {
    fprintf(stderr, "Error: no port given.\n");
    goto main_einval;
  }

  /* split user name */
  opt = strlen(user) + sizeof(char);
  if((utbuf = malloc(opt)) == NULL) {
    fprintf(stderr, "Error: not enough memory\n");
    return ENOMEM;
  } else {
    memset(utbuf, '\0', opt);
    memcpy(utbuf, user, opt - sizeof(char));
    if(ejda_username_split(utbuf, &m.user, &m.domain)) {
      fprintf(stderr, "Error: username is invalid.\n");
      free(utbuf);
      return EINVAL;
    }
  }

  /* TLS stuff, init and connect */
  ssllib_init();

  if(verbose) {
    fprintf(stdout, "Connecting to TLS port %s:%s ... ", host, port);
    fflush(stdout);
  }

  if(tls_connect(host, port, &p)) {
    if(!verbose)
      fprintf(stderr, "Error: failed to connect TLS port.\n");
    else fprintf(stdout, " failed\n");
    opt = -1;
    goto mainmoi;
  }
  if(verbose) fprintf(stdout, " ok\n");
  opt = 0;

  /* SMTP check user */
  if(verbose) {
    fprintf(stdout, "SMTP check user u=%s:d=%s ... ", m.user, m.domain);
    fflush(stdout);
  }

  opt = smtp_checkuser(&p, &m, host);
  /* invert result */
  if(expected && opt) opt = 0;
  if(expected && !opt) opt = 1;

  if(opt) {
    if(!verbose) fprintf(stderr, "SMTP check user returns not expected result.\n");
    else fprintf(stdout, " failed\n");
    opt = -1;
  } else {
    opt = 0;
    if(verbose) fprintf(stdout, " ok\n");
  }

  tls_close(&p);

 mainmoi:
  ssllib_free();
  free(utbuf);

  return opt;
}


