/*
 * ejabberd external authentication program
 *
 * (c) Alexander Vdolainen 2013, 2018, 2019, 2021 <alex@vapaa.xyz>
 *
 * this is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * this is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/select.h>
#include <errno.h>
#include <netdb.h>
#include <string.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#include <tlsport.h>

int ssllib_init(void)
{
  /* init the ssl library internl stuff */
  SSL_load_error_strings();
  SSL_library_init();

  return 0;
}

int ssllib_free(void)
{
  ERR_free_strings();

  EVP_cleanup();
  return 0;
}

int tls_connect(const char *hostname, const char *service, struct tlsport *p)
{
  struct addrinfo *host = NULL, *rh = NULL;
  struct addrinfo hint;
  int r = 0, sock, sel_read = 0, sel_write = 0;
  SSL_CTX *sslctx;
  SSL *sslc;
  fd_set readfd, writefd;

  /* get host by hostname first */
  /* set hints for search */
  memset(&hint, 0, sizeof(struct addrinfo));
  hint.ai_family = AF_UNSPEC; /* ipv4 and ipv6 support from the first step */
  hint.ai_socktype = SOCK_STREAM;
  r = getaddrinfo(hostname, service, &hint, &host);
  if(r) return -1; /* TODO: error report */

  for(rh = host; rh != NULL; rh = rh->ai_next) {
    sock = socket(rh->ai_family, rh->ai_socktype, rh->ai_protocol);
    if(sock == -1) continue;
    if(connect(sock, rh->ai_addr, rh->ai_addrlen) != -1) break; /* got it */
    else close(sock);
  }

  if(!rh) {
    freeaddrinfo(host);
    return -1; /* TODO: error report */
  } else freeaddrinfo(host);

  /* time to do a tls handshake */
  sslctx = SSL_CTX_new(SSLv23_client_method());
  if(!sslctx) {
  ssllib_error0:
    close(sock);
    return -1; /* TODO: error report */
  }
  /* disable old insecure SSLv3 */
  SSL_CTX_set_options(sslctx, SSL_OP_NO_SSLv3);
  /* create a new ssl connection context */
  sslc = SSL_new(sslctx);
  if(!sslc) {
  ssllib_error1:
    SSL_CTX_free(sslctx); /* TODO: error report */
    goto ssllib_error0;
  } else {
    SSL_set_fd(sslc, sock);
    SSL_set_connect_state(sslc); /* tls connection is going from client */
  }

 ssllib_connect:
  sel_read = sel_write = 0;
  do {
    r = SSL_connect(sslc);
    switch(SSL_get_error(sslc, r)) {
    case SSL_ERROR_NONE:
      r = SSL_ERROR_NONE; /* connected! */
      p->fd = sock;
      p->ssl = sslc;
      p->sslctx = sslctx;
      return 0;
      break;
    case SSL_ERROR_WANT_READ:
      sel_read = 1; /* is required to wait during read */ break;
    case SSL_ERROR_WANT_WRITE:
      sel_write = 1; /* is required to wait during write */ break;
    case SSL_ERROR_SYSCALL:
      if(errno == EAGAIN || errno == EINTR) break;
    case SSL_ERROR_SSL:
    case SSL_ERROR_ZERO_RETURN: /* expected close */
    default:
      SSL_free(sslc); /* TODO: error report */
      goto ssllib_error1;
      break;
    }
  } while(SSL_pending(sslc) && !sel_read);

  /* wait to read */
  if(r != SSL_ERROR_NONE && sel_read) {
    FD_ZERO(&readfd);
    FD_SET(sock, &readfd);

  read_select:
    r = select(sock + 1, &readfd, NULL, NULL, NULL);
    if(r <= 0) {
      if(errno == EAGAIN || errno == EINTR) goto read_select;
      else goto ssllib_error2;
    } else if(FD_ISSET(sock, &readfd)) goto ssllib_connect;
  }

  /* wait to write */
  if(r != SSL_ERROR_NONE && sel_write) {
    FD_ZERO(&readfd);
    FD_ZERO(&writefd);
    FD_SET(sock, &readfd);
    FD_SET(sock, &writefd);

    r = select(sock + 1, &readfd, &writefd, NULL, NULL);
    if(r && FD_ISSET(sock, &writefd)) goto ssllib_connect;
  }

 ssllib_error2:
  SSL_free(sslc);
  SSL_CTX_free(sslctx);
  close(sock);

  return -1;
}

int tls_close(struct tlsport *p)
{
  int r, e = 0;

  do {
    r = SSL_shutdown(p->ssl);
    if(r == 1) break;

    switch(SSL_get_error(p->ssl, r)) {
    case SSL_ERROR_WANT_READ:
    case SSL_ERROR_WANT_WRITE:
      break;
    case SSL_ERROR_SYSCALL:
      if(errno == EAGAIN || errno == EINTR) break;
    case SSL_ERROR_SSL:
    case SSL_ERROR_ZERO_RETURN: /* expected close */
    default:
      e = -1;
      break;
    }
  } while(SSL_pending(p->ssl) && !e);

  SSL_free(p->ssl);
  SSL_CTX_free(p->sslctx);
  close(p->fd);

  return e;
}

ssize_t tls_io(struct tlsport *p, void *buf, size_t bf_size, tls_io_dir_t dr)
{
  int sel_read, sel_write;
  int r, w = 0;
  fd_set readfd, writefd;

  if(!p || !buf) {
    errno = EINVAL;
    return -1;
  }

  if(!bf_size) {
    errno = EINVAL;
    return -1;
  }

 ssllib_io:
  sel_read = sel_write = 0;
  do {
    switch(dr) {
    case TLSIO_READ:
      r = SSL_read(p->ssl, buf, bf_size);
      if(r > 0) return r;
      break;
    case TLSIO_WRITE:
      r = SSL_write(p->ssl, (char *)buf + sizeof(char)*w, bf_size - w);
      if(r == (bf_size - w)) return bf_size;
      else if(r > 0) w += r;
      break;
    default:
      errno = EINVAL;
      return -1;
    }

    switch(SSL_get_error(p->ssl, r)) {
    case SSL_ERROR_NONE:
      break;
    case SSL_ERROR_WANT_READ:
      sel_read = 1; /* is required to wait during read */ break;
    case SSL_ERROR_WANT_WRITE:
      sel_write = 1; /* is required to wait during write */ break;
    case SSL_ERROR_WANT_CONNECT:
    case SSL_ERROR_WANT_ACCEPT:
      break;
    case SSL_ERROR_SYSCALL:
      if(errno == EAGAIN || errno == EINTR) break;
    case SSL_ERROR_SSL:
    case SSL_ERROR_ZERO_RETURN: /* expected close */
    default:
      return -1;
      break;
    }
  } while(SSL_pending(p->ssl) && !sel_read);

  /* wait to read */
  if(r != SSL_ERROR_NONE && sel_read) {
    FD_ZERO(&readfd);
    FD_SET(p->fd, &readfd);

  ioread_select:
    r = select(p->fd + 1, &readfd, NULL, NULL, NULL);
    if(r <= 0) {
      if(errno == EAGAIN || errno == EINTR) goto ioread_select;
      else return -1;
    } else if(FD_ISSET(p->fd, &readfd)) goto ssllib_io;
  }

  /* wait to write */
  if(r != SSL_ERROR_NONE && sel_write) {
    FD_ZERO(&readfd);
    FD_ZERO(&writefd);
    FD_SET(p->fd, &readfd);
    FD_SET(p->fd, &writefd);

    r = select(p->fd + 1, &readfd, &writefd, NULL, NULL);
    if(r && FD_ISSET(p->fd, &writefd)) goto ssllib_io;
  }

  return 0;
}
